"""Python package version query tools."""

import logging
import json
import pathlib

from .version import Version

_LOG = logging.getLogger(__name__)


def query_metadata_json(path: pathlib.Path) -> Version:
    """Get version from metadata.json file."""
    with path.open('r') as metadata_file:
        metadata = json.load(metadata_file)
    version_str = metadata['version']
    return Version.from_str(version_str)


def query_pkg_info(path: pathlib.Path) -> Version:
    """Get version from PKG-INFO file."""
    with path.open('r') as pkginfo_file:
        for line in pkginfo_file:
            if line.startswith('Version:'):
                version_str = line.replace('Version:', '').strip()
                return Version.from_str(version_str)
    raise ValueError(path)


def query_package_folder(path: pathlib.Path, search_parent_directories: bool = False) -> Version:
    """Get version from Python package folder."""
    global_metadata_json_paths, global_pkg_info_paths = [], []
    if path.joinpath('pyproject.toml').exists() or path.joinpath('setup.py').exists():
        metadata_json_paths = list(path.glob('*.dist-info/metadata.json'))
        pkg_info_paths = list(path.glob('*.egg-info/PKG-INFO'))
        pkg_info_paths += list(path.glob('*.dist-info/METADATA'))
        if len(metadata_json_paths) == 1 and not pkg_info_paths:
            return query_metadata_json(metadata_json_paths[0])
        if not metadata_json_paths and len(pkg_info_paths) == 1:
            return query_pkg_info(pkg_info_paths[0])
        _LOG.debug(
            'in %s found pyproject.toml or setup.py, as well as'
            ' %i JSON metadata: %s and %i PKG-INFO metadata: %s'
            ' - unable to infer package metadata, continuing search',
            path, len(metadata_json_paths), metadata_json_paths, len(pkg_info_paths),
            pkg_info_paths)
        global_metadata_json_paths.extend(metadata_json_paths)
        global_pkg_info_paths.extend(pkg_info_paths)
    paths = [path]
    if search_parent_directories:
        paths += path.parents
    for pth in paths:
        metadata_json_paths = list(pth.parent.glob(f'{pth.name}*.dist-info/metadata.json'))
        pkg_info_paths = list(pth.parent.glob(f'{pth.name}*.egg-info/PKG-INFO'))
        pkg_info_paths += list(pth.parent.glob(f'{pth.name}*.dist-info/METADATA'))
        if len(metadata_json_paths) == 1 and not pkg_info_paths:
            return query_metadata_json(metadata_json_paths[0])
        if not metadata_json_paths and len(pkg_info_paths) == 1:
            return query_pkg_info(pkg_info_paths[0])
        _LOG.debug(
            'in %s found %i JSON metadata: %s and %i PKG-INFO metadata: %s'
            ' - unable to infer package metadata, continuing search',
            pth, len(metadata_json_paths), metadata_json_paths, len(pkg_info_paths), pkg_info_paths)
        global_metadata_json_paths.extend(metadata_json_paths)
        global_pkg_info_paths.extend(pkg_info_paths)
    raise ValueError(
        f'unable to infer package metadata from the following paths {paths} '
        f'- found {len(global_metadata_json_paths)} JSON metadata: {global_metadata_json_paths}'
        f' and {len(global_pkg_info_paths)} PKG-INFO metadata: {global_pkg_info_paths}')



"""Tests for packaging."""

from version_query import predict_version_str

import boilerplates.packaging_tests

VERSION = predict_version_str()


class Tests(boilerplates.packaging_tests.PackagingTests):

    version = VERSION



"""Functions for parsing version strings into their components and verifying the format."""

import logging
import typing as t

from . import patterns

_LOG = logging.getLogger(__name__)


def parse_release_str(release: str) -> tuple:
    """Parse a release string into major, minor, and patch version numbers."""
    match = patterns.RELEASE.fullmatch(release)
    assert match is not None
    major_match = match.group('major')
    assert major_match is not None
    major = int(major_match)
    minor_match = match.group('minor')
    if minor_match is not None:
        minor = int(minor_match)
    else:
        minor = None
    patch_match = match.group('patch')
    if patch_match is not None:
        patch = int(patch_match)
    else:
        patch = None
    return major, minor, patch


def parse_pre_release_str(pre_release: str) -> t.Sequence[
        t.Tuple[t.Optional[str], t.Optional[str], t.Optional[int]]]:
    """Parse a pre-release string into a sequence of tuples."""
    parts = patterns.PRE_RELEASE.findall(pre_release)
    _LOG.debug('parsed pre-release string %s into %s', repr(pre_release), parts)
    tuples = []
    for part in parts:
        match = patterns.PRE_RELEASE_PART.fullmatch(part)
        assert match is not None
        pre_patch_match = match.group('prepatch')
        if pre_patch_match is not None:
            pre_patch = int(pre_patch_match)
        else:
            pre_patch = None
        tuples.append((match.group('preseparator'), match.group('pretype'), pre_patch))
    return tuples


def parse_local_str(local: str) -> tuple:
    """Parse a local version suffix string into a sequence."""
    match = patterns.LOCAL.fullmatch(local)
    assert match is not None
    return tuple(_ for _ in match.groups() if _ is not None)



"""Boilerplate helpful for testing in a context of a synthetic git repository."""

import logging
import pathlib
import tempfile
import typing as t
import unittest

import git

_LOG = logging.getLogger(__name__)


class GitRepoTests(unittest.TestCase):
    """Provide several utility properties and methods named repo_* and git_*."""

    repo: t.Optional[git.Repo] = None
    repo_path: t.Optional[pathlib.Path] = None

    def setUp(self):
        self._tmpdir = tempfile.TemporaryDirectory()  # pylint: disable = consider-using-with
        self.repo_path = pathlib.Path(self._tmpdir.name)
        self.assertTrue(self.repo_path.is_dir())
        self.repo = None
        self._repo_files = []

    def tearDown(self):
        for path in self._repo_files:
            if path.is_file():
                path.unlink()
        if self.repo is not None:
            self.assertIsInstance(self.repo, git.Repo)
            self.repo.close()
            self.repo = None
        self._tmpdir.cleanup()
        self._tmpdir = None

    @property
    def repo_head_hexsha(self) -> str:
        self.assertIsInstance(self.repo, git.Repo)
        assert isinstance(self.repo, git.Repo), type(self.repo)
        return self.repo.head.commit.hexsha[:8]

    def git_init(self) -> git.Repo:
        """Initialize a git repository in the temporary folder."""
        self.repo = git.Repo.init(str(self.repo_path))
        self.repo.git.config('user.email', 'you@example.com')
        self.repo.git.config('user.name', 'Your Name')
        return self.repo

    def git_clone(self, remote_name: str, url: str) -> git.Repo:
        """Clone a git repository into the temporary folder."""
        self.repo = git.Repo.clone_from(url, str(self.repo_path), origin=remote_name)
        self.assertIsInstance(self.repo, git.Repo)
        self.repo.git.config('user.email', 'you@example.com')
        self.repo.git.config('user.name', 'Your Name')
        return self.repo

    def git_commit_new_file(self) -> pathlib.Path:
        """Create a new file and commit it."""
        with tempfile.NamedTemporaryFile('w', dir=str(self.repo_path), delete=False) as repo_file:
            repo_file.write('spam spam lovely spam\n')
            path = pathlib.Path(repo_file.name)
        self.assertIsInstance(self.repo, git.Repo)
        assert isinstance(self.repo, git.Repo), type(self.repo)
        self.repo.index.add([path.name])
        self.repo.index.commit(f'created file {path}')
        _LOG.debug('commited file %s as %s', path, self.repo_head_hexsha)
        self._repo_files.append(path)
        return path

    def git_modify_file(self, path: pathlib.Path, add: bool = False, commit: bool = False) -> None:
        """Modify an existing file."""
        self.assertIsInstance(self.repo, git.Repo)
        assert isinstance(self.repo, git.Repo), type(self.repo)
        self.assertIsInstance(path, pathlib.Path)
        self.assertTrue(path.is_file())
        with path.open('a') as repo_file:
            repo_file.write('spam eggs ham\n')
        if add or commit:
            self.repo.index.add([path.name])
        if commit:
            self.repo.index.commit(f'modified file {path}')



"""Boilerplate useful when creating a command-line interface."""

import argparse
import logging
import sys
import textwrap
import typing as t

import argcomplete


class ArgumentDefaultsAndRawDescriptionHelpFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    """Inherit from ArgumentDefaultsHelpFormatter and RawDescriptionHelpFormatter."""


def make_copyright_notice(
        year_from: int, year_to: t.Optional[int] = None, author: str = 'Mateusz Bysiek',
        license_name: str = 'Apache License 2.0', url: t.Optional[str] = None) -> str:
    """Assemble a copyright notice like "Copyright YYYY by Author(s). License Name. http://url/"."""
    if year_to is None or year_to == year_from:
        years = str(year_from)
    else:
        years = f'{year_from}-{year_to}'
    return f'Copyright {years} by {author}. {license_name}. {"" if url is None else url}'.rstrip()


def add_version_option(parser: argparse.ArgumentParser, version: str):
    """Add --version option to a given parser."""
    parser.add_argument(
        '--version', action='version', version=f'{parser.prog} {version}, Python {sys.version}')


_LOGGING_LEVEL_MIN = logging.CRITICAL
_LOGGING_LEVEL_DEFAULT = logging.WARNING
_LOGGING_LEVEL_MAX = logging.DEBUG - 10


def verbosity_level_to_logging_level(verbosity: int) -> int:
    """Convert internal verbosity level to logging level from logging library."""
    return _LOGGING_LEVEL_MIN - verbosity * 10


def logging_level_to_verbosity_level(logging_level: int) -> int:
    """Convert logging level from logging library to internal verbosity level."""
    return (_LOGGING_LEVEL_MIN - logging_level) // 10


_VERBOSITY_MIN = logging_level_to_verbosity_level(_LOGGING_LEVEL_MIN)
_VERBOSITY_DEFAULT = logging_level_to_verbosity_level(_LOGGING_LEVEL_DEFAULT)
_VERBOSITY_MAX = logging_level_to_verbosity_level(_LOGGING_LEVEL_MAX)

_VERBOSE_MAX_COUNT = _VERBOSITY_MAX - _VERBOSITY_DEFAULT
_QUIET_MAX_COUNT = _VERBOSITY_DEFAULT - _VERBOSITY_MIN


def add_verbosity_group(parser: argparse.ArgumentParser) -> 'argparse._MutuallyExclusiveGroup':
    """Add parser option group for controlling application verbosity."""
    verbosity_group = parser.add_mutually_exclusive_group(required=False)
    verbosity_group.add_argument(
        '--verbose', '-v', action='count',
        help=f'''be more verbose than by default (repeat up to {_VERBOSE_MAX_COUNT} times
        for stronger effect)''')
    verbosity_group.add_argument(
        '--quiet', '-q', action='count',
        help=f'''be more quiet than by default (repeat up to {_QUIET_MAX_COUNT} times
        for stronger effect)''')
    verbosity_arg = verbosity_group.add_argument(
        '--verbosity', metavar='LEVEL', type=int, default=_VERBOSITY_DEFAULT,
        help=f'set verbosity level explicitly (normally from {_VERBOSITY_MIN} to {_VERBOSITY_MAX})')
    verbosity_arg.completer = argcomplete.completers.ChoicesCompleter(  # type: ignore
        choices=list(range(_VERBOSITY_MIN, _VERBOSITY_MAX)))
    return verbosity_group


def get_verbosity_level(parsed_args: argparse.Namespace) -> int:
    """Get verbosity level from parsed arguments after using add_verbosity_group() on a parser."""
    level = parsed_args.verbosity
    if parsed_args.verbose is not None:
        if parsed_args.verbose > _VERBOSE_MAX_COUNT:
            raise ValueError('too many repetitions of --verbose/-v')
        level += parsed_args.verbose
    if parsed_args.quiet is not None:
        if parsed_args.quiet > _QUIET_MAX_COUNT:
            raise ValueError('too many repetitions of --quiet/-q')
        level -= parsed_args.quiet
    return level


def get_logging_level(parsed_args: argparse.Namespace) -> int:
    """Get logging level equivalent to the verbosity from parsed arguments.

    This function only works after using add_verbosity_group().
    """
    return verbosity_level_to_logging_level(get_verbosity_level(parsed_args))


def dedent_except_first_line(text: str) -> str:
    """Dedent all lines of text except the 1st, using textwrap.dedent()."""
    try:
        newline = text.index('\n') + 1
    except ValueError:
        return text
    return text[:newline] + textwrap.dedent(text[newline:])



"""Initialization of version_query package."""

__all__ = ['VersionComponent', 'Version',
           'query_folder', 'query_caller', 'query_version_str', 'predict_git_repo',
           'predict_caller', 'predict_version_str']

from .version import VersionComponent, Version
from .query import query_folder, query_caller, query_version_str
from .git_query import predict_git_repo
from .query import predict_caller, predict_version_str



"""Boilerplate for integrating Sentry into the project."""

import logging
import os
import typing as t

import sentry_sdk
import sentry_sdk.integrations
import sentry_sdk.integrations.argv
import sentry_sdk.integrations.excepthook
import sentry_sdk.integrations.logging
import sentry_sdk.integrations.modules
import sentry_sdk.integrations.pure_eval
import sentry_sdk.integrations.stdlib
import sentry_sdk.integrations.threading

_LOG = logging.getLogger(__name__)


class Sentry:
    """Sentry configuration.

    For parameters 'dsn', 'release', 'environment', 'default_integrations', 'debug',
    'attach_stacktrace', 'shutdown_timeout', 'traces_sample_rate' and 'profiles_sample_rate'
    the value is taken from the environment variable if it is set,
    otherwise from the class attribute if it is set.
    For those parameters, the expected name of the environment variable name is 'SENTRY_'
    followed by the parameter name in upper case.

    The class attribute tags, if set, will be added to the global scope of the
    current Sentry SDK.
    """

    dsn: str
    """Sentry DSN (Data Source Name) used to identify the project in Sentry."""

    release: str
    """Release version of your application."""

    environment: str
    """Environment name (e.g., 'production', 'staging', 'development')."""

    integrations: t.List[sentry_sdk.integrations.Integration] = [
        sentry_sdk.integrations.argv.ArgvIntegration(),
        sentry_sdk.integrations.excepthook.ExcepthookIntegration(always_run=False),
        sentry_sdk.integrations.logging.LoggingIntegration(
            level=logging.INFO, event_level=logging.ERROR),
        sentry_sdk.integrations.modules.ModulesIntegration(),
        sentry_sdk.integrations.pure_eval.PureEvalIntegration(),
        sentry_sdk.integrations.stdlib.StdlibIntegration(),
        sentry_sdk.integrations.threading.ThreadingIntegration()
    ]
    """Sentry SDK integrations to enable.

    To configure integrations that are enabled by default, also add them to this list.
    """

    default_integrations: bool = True
    """If True, enable Sentry's default integrations besides the ones set via 'integrations'."""

    debug: bool = False
    """If True, enable debug mode for the Sentry SDK."""

    attach_stacktrace: bool = False
    """If True, attach stacktraces to messages."""

    shutdown_timeout: float = 2.0
    """Time in seconds to wait for pending events to be sent before shutdown."""

    send_default_pii: t.Optional[bool] = False
    """If True, send personally identifiable information."""

    traces_sample_rate: float = 1.0
    """Ratio of transactions to sample (0.0 to 1.0)."""

    profiles_sample_rate: float = 1.0
    """Ratio of transactions to profile (0.0 to 1.0)."""

    tags: t.Dict[str, str] = {}
    """Tags to be added to all events."""

    @classmethod
    def _get_str_param(cls, param_name: str) -> t.Optional[str]:
        """Get a string parameter value by checking envvar first.

        Works only for 'dsn', 'release' or 'environment' parameters.

        :param param_name: name of the parameter to get
        :return: value of the parameter
        """
        assert param_name in ('dsn', 'release', 'environment'), param_name
        return os.environ.get(f'SENTRY_{param_name.upper()}', getattr(cls, param_name, None))

    @classmethod
    def _get_float_param(cls, param_name: str) -> float:
        assert param_name in (
            'shutdown_timeout', 'traces_sample_rate', 'profiles_sample_rate'), param_name
        if f'SENTRY_{param_name.upper()}' in os.environ:
            return float(os.environ[f'SENTRY_{param_name.upper()}'])
        return getattr(cls, param_name)

    @classmethod
    def _get_bool_param(cls, param_name: str) -> bool:
        assert param_name in ('default_integrations', 'debug', 'attach_stacktrace'), param_name
        if f'SENTRY_{param_name.upper()}' in os.environ:
            return os.environ[f'SENTRY_{param_name.upper()}'].lower() in {'true', 'yes', 'on', '1'}
        return getattr(cls, param_name)

    @classmethod
    def is_dsn_set(cls) -> bool:
        """Check if Sentry DSN parameter is set, thus if Sentry SDK should be initialised or not."""
        dsn = cls._get_str_param('dsn')
        return dsn is not None and len(dsn) > 0

    @classmethod
    def init(cls, *args, **kwargs) -> None:
        """Initialise Sentry SDK."""
        if not cls.is_dsn_set():
            _LOG.info('Sentry DSN is not set, skipping Sentry SDK initialisation')
            return
        assert not args, \
            'positional arguments are not supported, instead please use class attributes,' \
            ' environment variables and for parameters which cannot be set via either' \
            ' please use keyword arguments'
        sentry_sdk.init(
            dsn=cls._get_str_param('dsn'),
            release=cls._get_str_param('release'),
            environment=cls._get_str_param('environment'),
            integrations=cls.integrations,
            traces_sample_rate=cls._get_float_param('traces_sample_rate'),
            profiles_sample_rate=cls._get_float_param('profiles_sample_rate'),
            enable_tracing=cls._get_float_param('profiles_sample_rate') > 0,
            debug=cls._get_bool_param('debug'),
            attach_stacktrace=cls._get_bool_param('attach_stacktrace'),
            shutdown_timeout=cls._get_float_param('shutdown_timeout'),
            send_default_pii=cls.send_default_pii,
            default_integrations=cls._get_bool_param('default_integrations'),
            **kwargs)

        sentry_sdk.set_tags(cls.tags)



"""Initialization of tests of boilerplates package."""

import logging

import boilerplates.logging


class TestsLogging(boilerplates.logging.Logging):
    """Logging configuration for tests."""

    packages = ['boilerplates']


TestsLogging.configure()

logging.getLogger('boilerplates.bundled_version_query').setLevel(logging.WARNING)



"""Tests for the CLI boilerplate."""

import argparse
import contextlib
import logging
import os
import pathlib
import unittest

import boilerplates.cli


class UnitTests(unittest.TestCase):
    """Test basic functionalities of the CLI boilerplate."""

    def test_copyright_notice(self):
        notice = boilerplates.cli.make_copyright_notice(2020)
        self.assertTrue(notice.startswith('Copyright 2020 '), msg=notice)
        notice = boilerplates.cli.make_copyright_notice(2020, 2020)
        self.assertTrue(notice.startswith('Copyright 2020 '), msg=notice)
        notice = boilerplates.cli.make_copyright_notice(2020, 2023)
        self.assertTrue(notice.startswith('Copyright 2020-2023 '), msg=notice)

    def test_version_option(self):
        parser = argparse.ArgumentParser('boilerplates_cli_test')
        boilerplates.cli.add_version_option(parser, '1.0.0')
        with pathlib.Path(os.devnull).open('w', encoding='utf-8') as devnull:
            with contextlib.redirect_stderr(devnull):
                with self.assertRaises(SystemExit):
                    parser.parse_args(['--version'])

    def test_logging_level_conversion(self):
        for logging_level in (
                logging.NOTSET, logging.DEBUG, logging.WARNING, logging.ERROR, logging.CRITICAL):
            with self.subTest(logging_level=logging_level):
                verbosity = boilerplates.cli.logging_level_to_verbosity_level(logging_level)
                self.assertGreaterEqual(
                    verbosity,
                    boilerplates.cli._VERBOSITY_MIN)  # pylint: disable = protected-access
                self.assertLessEqual(
                    verbosity,
                    boilerplates.cli._VERBOSITY_MAX)  # pylint: disable = protected-access
                converted_logging_level = \
                    boilerplates.cli.verbosity_level_to_logging_level(verbosity)
                self.assertEqual(converted_logging_level, logging_level)

    def test_verbosity_level_conversion(self):
        for verbosity in range(-10, 10):
            with self.subTest(verbosity=verbosity):
                logging_level = boilerplates.cli.verbosity_level_to_logging_level(verbosity)
                converted_verbosity = \
                    boilerplates.cli.logging_level_to_verbosity_level(logging_level)
                self.assertEqual(converted_verbosity, verbosity)

    def test_verbosity_default(self):
        parser = argparse.ArgumentParser()
        boilerplates.cli.add_verbosity_group(parser)
        parsed_args = parser.parse_args([])
        verbosity = boilerplates.cli.get_verbosity_level(parsed_args)
        self.assertEqual(
            verbosity, boilerplates.cli._VERBOSITY_DEFAULT)  # pylint: disable = protected-access

    def test_verbosity_by_level(self):
        parser = argparse.ArgumentParser()
        boilerplates.cli.add_verbosity_group(parser)
        for verbosity in range(
                boilerplates.cli._VERBOSITY_MIN,  # pylint: disable = protected-access
                boilerplates.cli._VERBOSITY_MAX):  # pylint: disable = protected-access
            parsed_args = parser.parse_args([f'--verbosity={verbosity}'])
            parsed_verbosity = boilerplates.cli.get_verbosity_level(parsed_args)
            self.assertEqual(parsed_verbosity, verbosity)
            logging_level = boilerplates.cli.get_logging_level(parsed_args)
            self.assertEqual(
                logging_level, boilerplates.cli.verbosity_level_to_logging_level(verbosity))

    def test_verbosity_by_flags(self):
        parser = argparse.ArgumentParser()
        boilerplates.cli.add_verbosity_group(parser)
        for flags, verbosity_change, in {
                ('-v',): 1,
                ('--verbose',): 1,
                ('-v', '-v'): 2,
                ('-vv',): 2,
                ('-q',): -1,
                ('--quiet',): -1,
                ('-q', '-q'): -2,
                ('-qq',): -2
                }.items():
            parsed_args = parser.parse_args(flags)
            verbosity = boilerplates.cli.get_verbosity_level(parsed_args)
            # pylint: disable = protected-access
            self.assertEqual(verbosity, boilerplates.cli._VERBOSITY_DEFAULT + verbosity_change)

    def test_too_many_verbosity_flags(self):
        parser = argparse.ArgumentParser()
        boilerplates.cli.add_verbosity_group(parser)
        # pylint: disable = protected-access
        for flags in (
                [f'-{"v" * (boilerplates.cli._VERBOSE_MAX_COUNT + 1)}'],
                [f'-{"v" * (boilerplates.cli._VERBOSE_MAX_COUNT + 2)}'],
                [f'-{"q" * (boilerplates.cli._QUIET_MAX_COUNT + 1)}'],
                [f'-{"q" * (boilerplates.cli._QUIET_MAX_COUNT + 2)}']):
            parsed_args = parser.parse_args(flags)
            with self.assertRaises(ValueError):
                boilerplates.cli.get_verbosity_level(parsed_args)

    def test_dedent_except_first_line(self):
        self.assertEqual(
            boilerplates.cli.dedent_except_first_line('  test'), '  test')
        self.assertEqual(
            boilerplates.cli.dedent_except_first_line('  test\n  test'), '  test\ntest')



"""Git repository version query tools."""

import datetime
import logging
import pathlib
import typing as t

import git

from .version import Version

_LOG = logging.getLogger(__name__)


def preprocess_git_version_tag(tag: str):
    """Remove a prefix from a version tag."""
    if tag.startswith('ver'):
        return tag[3:]
    if tag.startswith('v'):
        return tag[1:]
    if tag and tag[0] in ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9'):
        return tag
    raise ValueError(f'given tag "{tag}" does not appear to be a version tag')


def _git_version_tags(repo: git.Repo) -> t.Mapping[git.Tag, Version]:
    versions = {}
    for tag in repo.tags:
        try:
            tag_str = preprocess_git_version_tag(str(tag))
        except ValueError:
            _LOG.debug('%s: ignoring non-version tag %s', repo, tag)
            continue
        try:
            versions[tag] = Version.from_str(tag_str)
        except ValueError:
            # except packaging.version.InvalidVersion:
            _LOG.warning('%s: failed to convert %s to version', repo, tag_str)
            continue
    return versions


def _git_version_tag_commits(
        version_tags: t.Iterable[git.Tag]) -> t.Mapping[git.objects.Commit, t.Set[git.Tag]]:
    version_tag_commits: t.Dict[git.objects.Commit, t.Set[git.Tag]] = {}
    for tag in version_tags:
        _commit = tag.commit
        if _commit not in version_tag_commits:
            version_tag_commits[_commit] = set()
        version_tag_commits[_commit].add(tag)
    return version_tag_commits


def _latest_git_version_tag_on_branches(
        repo: git.Repo, assume_if_none: bool, commit: git.objects.Commit, commit_distance: int,
        skip_commits: t.Set[git.objects.Commit]) -> t.Union[int, t.Tuple[
            t.Optional[git.objects.Commit], t.Optional[git.TagReference], t.Optional[Version],
            int]]:
    _LOG.log(logging.NOTSET, 'entering %i branches...', len(commit.parents))
    results: t.List[t.Tuple[
        t.Optional[git.objects.Commit], t.Optional[git.TagReference], Version, int]] = []
    main_commit_distance = None
    for parent in commit.parents:
        try:
            result = _latest_git_version_tag(
                repo, assume_if_none, parent, commit_distance, skip_commits)
        except ValueError:
            continue
        if main_commit_distance is None:
            main_commit_distance = result[3]
        if result[2] is not None:
            results.append(result)  # type: ignore
    if not results:
        if main_commit_distance is None:
            raise ValueError(f'reached max commit distance {MAX_COMMIT_DISTANCE}'
                             f' with no version tags in repo {repo}')
        return main_commit_distance
    final_result = sorted(results, key=lambda _: _[2])[-1]
    _LOG.log(logging.NOTSET, 'result from %i branches is %s and %s',
             len(commit.parents), *final_result[1:3])
    return final_result


MAX_COMMIT_DISTANCE = 999


def _latest_git_version_tag(
        repo: git.Repo, assume_if_none: bool = False,
        base_commit: t.Optional[git.objects.Commit] = None, commit_distance: int = 0,
        skip_commits: t.Optional[t.Set[git.objects.Commit]] = None) -> t.Tuple[
            t.Optional[git.objects.Commit], t.Optional[git.TagReference], t.Optional[Version], int]:
    """Return (commit, tag at that commit if any, latest version, distance from the version)."""
    version_tags = _git_version_tags(repo)
    version_tag_commits = _git_version_tag_commits(version_tags.keys())
    current_version_tags = {}
    commit = None
    if skip_commits is None:
        skip_commits = set()
    for commit in repo.iter_commits(rev=base_commit):
        if commit in skip_commits:
            return None, None, None, -1
        _LOG.log(logging.NOTSET, 'iterating over commit %s', commit)
        skip_commits.add(commit)
        if commit in version_tag_commits:
            current_tags = version_tag_commits[commit]
            current_version_tags = {tag: version for tag, version in version_tags.items()
                                    if tag in current_tags}
            _LOG.log(logging.NOTSET, 'found version data %s', current_version_tags)
            break
        if commit_distance >= MAX_COMMIT_DISTANCE:
            raise ValueError(f'reached max commit distance {MAX_COMMIT_DISTANCE}'
                             f' with no version tags in repo {repo}')
        commit_distance += 1
        if len(commit.parents) <= 1:
            continue
        result = _latest_git_version_tag_on_branches(
            repo, assume_if_none, commit, commit_distance, skip_commits)
        if not isinstance(result, tuple):
            commit_distance = result  # main_commit_distance
            break
        return result
    if not current_version_tags:
        if assume_if_none:
            return commit, None, Version.from_str('0.1.0.dev0'), commit_distance
        raise ValueError(f'the given repo {repo} has no version tags')
    tag, version = sorted(current_version_tags.items(), key=lambda _: _[1])[-1]
    _LOG.log(logging.NOTSET, 'result is %s and %s', tag, version)
    return commit, tag, version, commit_distance


def _upcoming_git_version_tag(repo: git.Repo, ignore_untracked_files: bool = True) -> t.Tuple[
        t.Optional[git.objects.Commit], t.Optional[git.TagReference], t.Optional[Version], int,
        bool]:
    commit, tag, version, commit_distance = _latest_git_version_tag(repo, True)
    is_repo_dirty = repo.is_dirty(untracked_files=not ignore_untracked_files)
    return commit, tag, version, commit_distance, is_repo_dirty


def query_git_repo(repo_path: pathlib.Path, search_parent_directories: bool = True) -> Version:
    """Determine version from tags of a git repository."""
    _LOG.debug('looking for git repository in "%s"', repo_path)
    repo = git.Repo(str(repo_path), search_parent_directories=search_parent_directories)
    _LOG.debug('found git repository in "%s"', repo.working_dir)
    version = _latest_git_version_tag(repo)[2]
    assert isinstance(version, Version), version
    return version


def predict_git_repo(repo_path: pathlib.Path, search_parent_directories: bool = True) -> Version:
    """Predict version from tags, commit history and index status of git repository."""
    repo = git.Repo(str(repo_path), search_parent_directories=search_parent_directories)
    version, commit_distance, is_repo_dirty = _upcoming_git_version_tag(repo)[2:]
    assert isinstance(version, Version), version
    if commit_distance > 0:
        version.devel_increment(commit_distance)
        version.local = (f'git{repo.head.commit.hexsha[:8]}',)
    if is_repo_dirty:
        dt_ = f'dirty{datetime.datetime.strftime(datetime.datetime.now(), "%Y%m%d%H%M%S")}'
        if version.has_local:
            assert version.local is not None  # mypy needs this
            version.local = (*version.local, '.', dt_)
        else:
            version.local = (dt_,)
    return version



"""Tests for boilerplates.config module."""

import os
import pathlib
import tempfile
import unittest
import unittest.mock

import boilerplates.config


class UnitTests(unittest.TestCase):
    """Test basic functionalities of the config boilerplate."""

    def test_normalize_path(self):
        home_path = os.path.expanduser('~')
        self.assertEqual(boilerplates.config.normalize_path('~/folder'), f'{home_path}/folder')

        self.assertEqual(
            boilerplates.config.normalize_path(pathlib.Path('~', 'something')),
            pathlib.Path(home_path).joinpath('something'))

        envvar_value = 'my_custom_value'
        os.environ['MY_CUSTOM_VAR'] = envvar_value

        self.assertEqual(
            boilerplates.config.normalize_path(r'${MY_CUSTOM_VAR}/folder'),
            f'{envvar_value}/folder')

        self.assertEqual(
            boilerplates.config.normalize_path(pathlib.Path(r'${MY_CUSTOM_VAR}', 'something')),
            pathlib.Path(envvar_value).joinpath('something'))

    def test_initialize_config_directory(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            with unittest.mock.patch.object(
                    boilerplates.config, 'CONFIGS_PATH', pathlib.Path(temp_dir)):
                self.assertTrue(boilerplates.config.CONFIGS_PATH.is_dir())
                boilerplates.config.initialize_config_directory('my_app')
                self.assertTrue(boilerplates.config.CONFIGS_PATH.joinpath('my_app').is_dir())

            with unittest.mock.patch.object(
                    boilerplates.config, 'CONFIGS_PATH', pathlib.Path(temp_dir, 'logs')):
                self.assertFalse(boilerplates.config.CONFIGS_PATH.exists())
                boilerplates.config.initialize_config_directory('my_app')
                self.assertTrue(boilerplates.config.CONFIGS_PATH.is_dir())
                self.assertTrue(boilerplates.config.CONFIGS_PATH.joinpath('my_app').is_dir())

            pathlib.Path(temp_dir, 'logging').mkdir()
            with unittest.mock.patch.object(
                    boilerplates.config, 'CONFIGS_PATH', pathlib.Path(temp_dir)):
                self.assertTrue(boilerplates.config.CONFIGS_PATH.joinpath('logging').is_dir())
                boilerplates.config.initialize_config_directory('logging')
                self.assertTrue(boilerplates.config.CONFIGS_PATH.joinpath('logging').is_dir())



"""Perform tests on and in synthetic git repositories."""

import pathlib

import boilerplates.git_repo_tests


class GitRepoSelfTests(boilerplates.git_repo_tests.GitRepoTests):
    """Check that GitRepoTests class works."""

    def test_typical(self):
        self.git_init()
        pth = self.git_commit_new_file()
        self.git_modify_file(pth)
        pth = self.git_commit_new_file()
        self.git_modify_file(pth, add=True)
        pth = self.git_commit_new_file()
        self.git_modify_file(pth, commit=True)

    def test_clone(self):
        path = pathlib.Path(__file__).resolve().parent.parent
        self.git_clone('origin', str(path))

    def test_no_repo(self):
        self.assertIsNone(self.repo)

    def test_cleanup_nonexisting(self):
        self.git_init()
        pth = self.git_commit_new_file()
        pth.unlink()
        self.assertFalse(pth.is_file())



"""Boilerplate useful to setup logging."""

import collections.abc
import datetime
import logging
import logging.config
import os
import pathlib
import platform
import typing as t

import colorlog

from .config import normalize_path

LOGS_PATHS = {
    'Linux': pathlib.Path('~', '.local', 'share'),
    'Darwin': pathlib.Path('~', 'Library', 'Logs'),
    'Windows': pathlib.Path('%LOCALAPPDATA%')}

LOGS_PATH = LOGS_PATHS[platform.system()]

DEFAULT_LOGGING_LEVEL_GLOBAL = logging.DEBUG
DEFAULT_LOGGING_LEVEL_PACKAGE = logging.DEBUG
DEFAULT_LOGGING_LEVEL_TEST = logging.DEBUG
DEFAULT_LOGGING_LEVEL_OTHER = logging.WARNING

LEVEL_ENVVAR_NAME = 'LOGGING_LEVEL'

DATETIME_FORMAT_DAILY = r'%Y%m%d'
DATETIME_FORMAT_PRECISE = r'%Y%m%d-%H%M%S'

LOG_FORMAT_BRIEF = r'{name} [{levelname}] {message}'
LOG_FORMAT_BRIEF_COLOURED = r'{name} [{log_color}{levelname}{reset}] {message}'
LOG_FORMAT_PRECISE = f'{{asctime}} {LOG_FORMAT_BRIEF}'
LOG_FORMAT_PRECISE_COLOURED = f'{{asctime}} {LOG_FORMAT_BRIEF_COLOURED}'


def logging_level_from_envvar(envvar: str, default: int = logging.WARNING) -> int:
    """Translate text envvar value into an integer corresponding to a logging level."""
    envvar_value = os.environ.get(envvar)
    if envvar_value is None:
        return default
    envvar_value = envvar_value.upper()
    if not hasattr(logging, envvar_value):
        try:
            return int(envvar_value)
        except ValueError:
            return default
    return getattr(logging, envvar_value)


def log_filename_basic(app_name: str) -> str:
    return f'{app_name}.log'


def log_filename_daily(app_name: str) -> str:
    timestamp = datetime.datetime.now().strftime(DATETIME_FORMAT_DAILY)
    return f'{app_name}_{timestamp}.log'


def log_filename_precise(app_name: str) -> str:
    timestamp = datetime.datetime.now().strftime(DATETIME_FORMAT_PRECISE)
    return f'{app_name}_{timestamp}.log'


class Logging:
    """Boilerplate to configure logging for an application."""

    packages: t.List[str]
    """List of packages that are considered in-package/local/internal."""

    directory: str
    """This must be set if enable_file is True."""

    filename: t.Optional[str] = None
    """Choose a custom log filename, instead of an auto-generated one."""

    enable_console: bool = True
    """Default True, you can disable console logging by setting this to False."""

    enable_file: bool = False
    """Default False, you can enable file logging by setting this to True."""

    level_global: int = DEFAULT_LOGGING_LEVEL_GLOBAL
    """Global logging cut-off filter.

    If set higher than any other level_* fields, it overrides then and filters all messages
    below the level.

    If set lower than some of other level_* fields, the values of respective fields still apply.
    """

    level_package: int = DEFAULT_LOGGING_LEVEL_PACKAGE
    """Logging level for in-package logging.

    This applies to the packages in the packages field.
    """

    level_test: int = DEFAULT_LOGGING_LEVEL_TEST
    """Logging level for tests.

    This applies to packages with name basically starting with 'test.*'.
    """

    level_other: int = DEFAULT_LOGGING_LEVEL_OTHER
    """Logging level for all other packages.

    This applies to the root logger, so to all packages not covered by level_package and level_test.
    """

    @classmethod
    def _log_absolute_path(cls) -> pathlib.Path:
        assert cls.directory is not None
        filename = log_filename_precise(cls.directory) if cls.filename is None else cls.filename
        return normalize_path(LOGS_PATH.joinpath(cls.directory, filename))

    @classmethod
    def _create_logs_folder(cls) -> None:
        assert cls.directory is not None
        logs_path = normalize_path(LOGS_PATH.joinpath(cls.directory))
        assert logs_path.parent.exists(), logs_path.parent
        logs_path.mkdir(parents=False, exist_ok=True)

    @classmethod
    def _set_default_logging_levels(cls) -> None:
        logging.getLogger().setLevel(cls.level_other)
        for package in getattr(cls, 'packages', []):
            logging.getLogger(package).setLevel(cls.level_package)
        logging.getLogger('test').setLevel(cls.level_test)

    @classmethod
    def configure_basic(cls) -> None:
        """Configure basic logging for an application.

        Basic logging is logging to the console with colored logging, or logging to a single file.
        """
        if cls.enable_console:
            assert not cls.enable_file
            cls._configure_basic_console()
        elif cls.enable_file:
            cls._create_logs_folder()
            logging.basicConfig(
                level=logging_level_from_envvar(LEVEL_ENVVAR_NAME, default=cls.level_global),
                filename=str(cls._log_absolute_path()))
        else:
            logging.basicConfig(
                level=logging_level_from_envvar(LEVEL_ENVVAR_NAME, default=cls.level_global))

        cls._set_default_logging_levels()

    @classmethod
    def _configure_basic_console(cls) -> None:
        """Configure basic logging to the console with colored logging."""
        handler = logging.StreamHandler()
        handler.setFormatter(colorlog.ColoredFormatter(LOG_FORMAT_BRIEF_COLOURED, style='{'))

        logging.basicConfig(
            level=logging_level_from_envvar(LEVEL_ENVVAR_NAME, default=cls.level_global),
            handlers=[handler])

    @classmethod
    def configure(cls) -> None:
        """Configure logging for an application."""
        logging_config: dict[str, t.Any] = {
            'formatters': {
                'console': {
                    '()': 'colorlog.ColoredFormatter',
                    'style': '{',
                    'format': LOG_FORMAT_PRECISE_COLOURED},
                'file': {
                    'style': '{',
                    'format': LOG_FORMAT_PRECISE}},
            'handlers': {},
            'root': {
                'handlers': [],
                'level': logging.NOTSET},
            'version': 1,
            'disable_existing_loggers': False}
        if cls.enable_console:
            logging_config['handlers']['console'] = {
                'class': 'logging.StreamHandler',
                'formatter': 'console',
                'level': logging_level_from_envvar(LEVEL_ENVVAR_NAME, default=cls.level_global),
                'stream': 'ext://sys.stdout'}
            logging_config['root']['handlers'].append('console')
        if cls.enable_file:
            cls._create_logs_folder()
            logging_config['handlers']['file'] = {
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'file',
                'level': logging.NOTSET,
                'filename': str(cls._log_absolute_path()),
                'maxBytes': 1 * 1024 * 1024,
                'backupCount': 10}
            logging_config['root']['handlers'].append('file')
        logging.config.dictConfig(logging_config)

        cls._set_default_logging_levels()


def unittest_verbosity() -> t.Optional[int]:
    """Retrieve the verbosity setting of the currently running unittest program.

    Return None if currently running program is not unittest.

    Default verbosity level is 1, 0 means quiet and 2 means verbose.
    """
    import inspect  # pylint: disable = import-outside-toplevel
    import unittest  # pylint: disable = import-outside-toplevel

    frame = inspect.currentframe()
    while frame:
        self_ = frame.f_locals.get('self')
        if isinstance(self_, unittest.TestProgram):
            return self_.verbosity
        frame = frame.f_back
    return None


class StreamToCall(t.IO[str]):
    """Redirect stream writes to a function call.

    Enable using logging instances as a file-like objects.
    Given a called_function, convert write(text) calls to called_function(text) calls.
    For example: StreamToCall(logging.warning) will redirect all writes to logging.warning().

    Example usage:

    logger = logging.getLogger(__name__)
    with contextlib.redirect_stdout(boilerplates.logging.StreamToCall(logger.info)):
        print('this will be logged at INFO level')

    Another example:

    logger = logging.getLogger(__name__)
    stream = boilerplates.logging.StreamToCall(logger.debug)
    print('this will be logged at DEBUG level', file=stream)
    """

    def __init__(self, called_function: collections.abc.Callable):
        assert callable(called_function), type(called_function)
        self._function = called_function

    def write(self, message: str, *args):
        """Redirect the write to the logging function."""
        while message.endswith('\r') or message.endswith('\n'):
            message = message[:-1]
        self._function(message, *args)
        return len(message)

    def flush(self):
        """Flush can be a no-op."""



"""Unit tests for Sentry boilerplate."""

import logging
import sys
import unittest
import unittest.mock

import boilerplates.sentry


class SentryTests(unittest.TestCase):

    def test_init(self):
        with self.assertLogs('boilerplates.sentry', logging.INFO) as context:
            boilerplates.sentry.Sentry.init()
        self.assertEqual(len(context.output), 1, msg=context.output)
        self.assertIn('skipping Sentry SDK initialisation', context.output[0])

    def test_init_without_dsn(self):
        class Sentry(boilerplates.sentry.Sentry):
            dsn = ''
        with self.assertLogs('boilerplates.sentry', logging.INFO) as context, \
                unittest.mock.patch('sentry_sdk.init') as sentry_sdk_init_mock:
            Sentry.init()
        sentry_sdk_init_mock.assert_not_called()
        self.assertEqual(len(context.output), 1, msg=context.output)
        self.assertIn('skipping Sentry SDK initialisation', context.output[0])

    @unittest.skipUnless(sys.version_info >= (3, 10), 'this test requires Python 3.10')
    def test_init_with_dsn(self):
        class Sentry(boilerplates.sentry.Sentry):
            dsn = 'https://spam@ham.ingest.sentry.io/eggs'
        with self.assertNoLogs('boilerplates.sentry', logging.INFO), \
                unittest.mock.patch('sentry_sdk.init') as sentry_sdk_init_mock:
            Sentry.init()
        sentry_sdk_init_mock.assert_called_once()

    @unittest.skipUnless(sys.version_info >= (3, 10), 'this test requires Python 3.10')
    def test_init_with_envvar(self):
        class Sentry(boilerplates.sentry.Sentry):
            pass
        environ_override = {'SENTRY_DSN': 'https://spam@ham.ingest.sentry.io/eggs'}
        with unittest.mock.patch.dict('os.environ', environ_override), \
                self.assertNoLogs('boilerplates.sentry', logging.INFO), \
                unittest.mock.patch('sentry_sdk.init') as sentry_sdk_init_mock:
            Sentry.init()
        sentry_sdk_init_mock.assert_called_once()

    @unittest.skipUnless(sys.version_info >= (3, 10), 'this test requires Python 3.10')
    def test_configure_with_envvars(self):
        class Sentry(boilerplates.sentry.Sentry):
            dsn = 'https://spam@ham.ingest.sentry.io/eggs'
        environ_override = {'SENTRY_PROFILES_SAMPLE_RATE': '0.42', 'SENTRY_DEBUG': 'True'}
        with unittest.mock.patch.dict('os.environ', environ_override), \
                self.assertNoLogs('boilerplates.sentry', logging.INFO), \
                unittest.mock.patch('sentry_sdk.init') as sentry_sdk_init_mock:
            Sentry.init()
        sentry_sdk_init_mock.assert_called_once()
        self.assertIn('profiles_sample_rate', sentry_sdk_init_mock.call_args.kwargs)
        self.assertEqual(sentry_sdk_init_mock.call_args.kwargs['profiles_sample_rate'], 0.42)
        self.assertIn('debug', sentry_sdk_init_mock.call_args.kwargs)
        self.assertTrue(sentry_sdk_init_mock.call_args.kwargs['debug'], 0.42)



"""Patterns for recognising parts of version strings and parsing them."""

import re

# pylint: disable = consider-using-f-string

_NUMBER = r'(?:0|[123456789][0123456789]*)'
# _SHA = r'[0123456789abcdef]+'
_LETTERS = r'(?:[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ]+)'
LETTERS = re.compile(_LETTERS)
_ALPHANUMERIC = r'(?:[0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ]+)'
ALPHANUMERIC = re.compile(_ALPHANUMERIC)
_SEP = r'(?:[\.-])'

_RELEASE_PARTS = \
    r'(?P<major>{n})(?:\.(?P<minor>{n}))?(?:\.(?P<patch>{n}))?'.format(n=_NUMBER)
RELEASE = re.compile(_RELEASE_PARTS)

_PRE_SEPARATOR = rf'(?P<preseparator>{_SEP})'
_PRE_TYPE = rf'(?P<pretype>{_LETTERS})'
_PRE_PATCH = rf'(?P<prepatch>{_NUMBER})'
_PRE_RELEASE_PART = rf'{_PRE_SEPARATOR}?{_PRE_TYPE}?{_PRE_PATCH}?'
PRE_RELEASE_PART = re.compile(_PRE_RELEASE_PART)
_PRE_RELEASE_PARTS = r'(?:{0}{2})|(?:{0}?{1}{2}?)'.format(_SEP, _LETTERS, _NUMBER)
PRE_RELEASE = re.compile(_PRE_RELEASE_PARTS)
# PRE_RELEASE_CHECK = re.compile(rf'(?:{_PRE_RELEASE_PARTS})+')

_LOCAL_SEPARATOR = rf'({_SEP})'
_LOCAL_PART = rf'({_ALPHANUMERIC})'
_LOCAL_PARTS = rf'\+{_LOCAL_PART}(?:{_LOCAL_SEPARATOR}{_LOCAL_PART})*'
LOCAL = re.compile(_LOCAL_PARTS)


_RELEASE = r'(?P<release>{n}(?:\.{n})?(?:\.{n})?)'.format(n=_NUMBER)
_PRE_RELEASE = r'(?P<prerelease>(?:(?:{0}{2})|(?:{0}?{1}{2}?))+)'.format(_SEP, _LETTERS, _NUMBER)
_LOCAL = r'(?P<local>\+{0}([\.-]{0})*)'.format(_ALPHANUMERIC)
# _NAMED_PARTS_COUNT = 3 + 3
_VERSION = rf'{_RELEASE}{_PRE_RELEASE}?{_LOCAL}?'
VERSION = re.compile(_VERSION)



"""Setup script for boilerplates package."""

import pathlib
import shutil
import tempfile

import git

import boilerplates.setup


def prepare_local_version_query():
    """Prepare local copy of version_query package to avoid circular dependency between packages."""
    here = pathlib.Path(__file__).parent
    module_path = here / 'boilerplates' / 'bundled_version_query'
    if module_path.exists():
        return
    with tempfile.TemporaryDirectory() as temporary_path:
        repo_path = pathlib.Path(temporary_path)
        print(f'ensuring version-query repository is available locally at "{repo_path}"')
        repo = git.Repo.clone_from('https://github.com/mbdevpl/version-query', repo_path)
        repo.git.checkout('v1.6.3')
        print(f'ensuring version_query module is available locally at "{module_path}"')
        shutil.copytree(repo_path / 'version_query', module_path, dirs_exist_ok=True)
    for filename in ('_version.py', '__main__.py', 'main.py'):
        module_path.joinpath(filename).unlink()


prepare_local_version_query()

# pylint: disable = wrong-import-position
from boilerplates.bundled_version_query import predict_version_str  # noqa: E402

VERSION = predict_version_str()


class Package(boilerplates.setup.Package):
    """Package metadata."""

    name = 'boilerplates'
    version = VERSION
    description = 'Various boilerplates used in almost all of my Python packages.'
    url = 'https://github.com/mbdevpl/python-boilerplates'
    author = 'Mateusz Bysiek, Stian Hanssen'
    maintainer = 'Mateusz Bysiek'
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: System :: Archiving :: Packaging',
        'Topic :: System :: Installation/Setup',
        'Topic :: System :: Logging',
        'Topic :: System :: Software Distribution',
        'Topic :: Utilities',
        'Typing :: Typed']
    keywords = ['git', 'logging', 'packaging', 'releasing']
    extras_require = {
        'setup': boilerplates.setup.parse_requirements('requirements_setup.txt'),
        'packaging-tests': boilerplates.setup.parse_requirements(
            'requirements_packaging_tests.txt'),
        'config': boilerplates.setup.parse_requirements('requirements_config.txt'),
        'logging': boilerplates.setup.parse_requirements('requirements_logging.txt'),
        'sentry': boilerplates.setup.parse_requirements('requirements_sentry.txt'),
        'cli': boilerplates.setup.parse_requirements('requirements_cli.txt'),
        'git-repo-tests': boilerplates.setup.parse_requirements('requirements_git_repo_tests.txt')}


if __name__ == '__main__':
    Package.setup()



"""Test definitions for package building."""

import contextlib
import importlib
import io
import logging
import os
import pathlib
import runpy
import subprocess
import sys
import tempfile
import types
import typing as t
import unittest

from .setup import find_version

_LOG = logging.getLogger(__name__)


def expand_args_by_globbing_items(
        *args: str, cwd: t.Optional[pathlib.Path] = None) -> t.Tuple[str, ...]:
    """Expand a list of glob expressions."""
    if cwd is None:
        cwd = pathlib.Path.cwd()
    expanded_args = []
    for arg in args:
        if '*' not in arg:
            expanded_args.append(arg)
            continue
        expanded_arg = [str(_.relative_to(cwd)) for _ in cwd.glob(arg)]
        assert expanded_arg, arg
        _LOG.debug('expanded arg "%s" to %s', arg, expanded_arg)
        expanded_args += expanded_arg
    _LOG.debug('expanded args to %s', expanded_args)
    return tuple(expanded_args)


def run_program(*args, glob: bool = False) -> None:
    """Run subprocess with given args. Use path globbing for each arg that contains an asterisk."""
    if glob:
        args = expand_args_by_globbing_items(*args)
    try:
        subprocess.run(args, check=True)
    except subprocess.CalledProcessError as err:
        raise AssertionError(f'execution of {args} failed') from err


def run_pip(*args, **kwargs) -> None:
    run_program(sys.executable, '-m', 'pip', *args, **kwargs)


def run_module(name: str, *args, run_name: str = '__main__') -> None:
    backup_sys_argv = sys.argv
    sys.argv = [name + '.py'] + list(args)
    runpy.run_module(name, run_name=run_name)
    sys.argv = backup_sys_argv


def import_module(name: str = 'setup') -> types.ModuleType:
    setup_module = importlib.import_module(name)
    return setup_module


def import_module_member(module_name: str, member_name: str) -> t.Any:
    module = import_module(module_name)
    return getattr(module, member_name)


def get_package_folder_name():
    """Attempt to guess the built package name."""
    name_from_setup = import_module_member('setup', 'Package').name.replace('-', '_')
    cwd = pathlib.Path.cwd()
    directories = [
        path for path in cwd.iterdir() if pathlib.Path(cwd, path).is_dir()
        and pathlib.Path(cwd, path, '__init__.py').is_file() and path.name != 'test']
    directory_names = {_.name for _ in directories}
    assert name_from_setup in directory_names, directories
    return name_from_setup


@unittest.skipUnless(os.environ.get('TEST_PACKAGING') or os.environ.get('CI'),
                     'skipping packaging tests for actual package')
class PackagingTests(unittest.TestCase):
    """Test if the boilerplate can actually create a valid package.

    Note:

    Tests that run setup script directly fail due to conflict between distutils used internally
    by setuptools and distutils used internally by pip. For more information, see:
    - https://github.com/pypa/setuptools/issues/3297
    - https://github.com/pypa/pip/issues/8761
    """

    pkg_name: t.Optional[str] = None
    version: t.Optional[str] = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.pkg_name = get_package_folder_name() if cls.pkg_name is None else cls.pkg_name
        cls.version = find_version(cls.pkg_name) if cls.version is None else cls.version

    def test_expand_args_in_cwd(self):
        expanded_args = expand_args_by_globbing_items('*.py')
        self.assertIn('setup.py', expanded_args)

    def test_expand_args_in_package_folder(self):
        expanded_args = expand_args_by_globbing_items('*.py', cwd=pathlib.Path(self.pkg_name))
        self.assertIn('__init__.py', expanded_args)

    @unittest.skipIf(sys.version_info < (3, 12), 'requires Python 3.12+')
    def test_setup_help(self):
        stdout_buffer = io.StringIO()
        with contextlib.redirect_stdout(stdout_buffer):
            run_module('setup', '--help')
        self.assertIn('usage: setup.py', stdout_buffer.getvalue())

    def test_build(self):
        run_program(sys.executable, '-m', 'build')
        self.assertTrue(os.path.isdir('dist'))

    def test_build_wheel(self):
        run_module('build', '--wheel')
        self.assertTrue(os.path.isdir('dist'))

    def test_build_wheel_no_isolation(self):
        run_module('build', '--wheel', '--no-isolation')
        self.assertTrue(os.path.isdir('dist'))

    def test_build_source(self):
        run_module('build', '--sdist')
        self.assertTrue(os.path.isdir('dist'))

    def test_build_source_no_isolation(self):
        run_module('build', '--sdist', '--no-isolation')
        self.assertTrue(os.path.isdir('dist'))

    def test_install_code(self):
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip('install', '--ignore-installed', '--prefix', temporary_folder, '.')
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_install_source_tar(self):
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip(
                'install', '--ignore-installed', '--prefix', temporary_folder,
                f'dist/*-{self.version}.tar.gz', glob=True)
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_install_wheel(self):
        with tempfile.TemporaryDirectory() as temporary_folder:
            run_pip(
                'install', '--ignore-installed', '--prefix', temporary_folder,
                f'dist/*-{self.version}-*.whl', glob=True)
        self.assertFalse(pathlib.Path(temporary_folder).exists())

    def test_pip_error(self):
        with self.assertRaises(AssertionError):
            run_pip('wrong_pip_command')

    @unittest.skipIf(sys.version_info < (3, 12), 'requires Python 3.12+')
    def test_setup_do_nothing_or_error(self):
        run_module('setup', 'wrong_setup_command', run_name='__not_main__')
        with self.assertRaises(SystemExit):
            run_module('setup', 'wrong_setup_command')



"""Tests for boilerplates.setup module."""

import itertools
import logging
import os
import pathlib
import tempfile
import unittest

import boilerplates.setup

_LOG = logging.getLogger(__name__)


CLASSIFIERS_LICENSES = (
    'License :: OSI Approved :: Python License (CNRI Python License)',
    'License :: OSI Approved :: Python Software Foundation License',
    'License :: Other/Proprietary License',
    'License :: Public Domain')

CLASSIFIERS_PYTHON_VERSIONS = tuple("""Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.0
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3 :: Only""".splitlines())

CLASSIFIERS_PYTHON_IMPLEMENTATIONS = tuple("""Programming Language :: Python :: Implementation
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: Jython
Programming Language :: Python :: Implementation :: PyPy
Programming Language :: Python :: Implementation :: Stackless""".splitlines())

CLASSIFIERS_VARIOUS = (
    'Framework :: IPython',
    'Topic :: Scientific/Engineering',
    'Topic :: Sociology',
    'Topic :: Security :: Cryptography',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Version Control :: Git',
    'Topic :: System',
    'Topic :: Utilities')

CLASSIFIERS_LICENSES_TUPLES = tuple((_,) for _ in CLASSIFIERS_LICENSES) + ((),)

CLASSIFIERS_PYTHON_VERSIONS_COMBINATIONS = tuple((_,) for _ in CLASSIFIERS_PYTHON_VERSIONS)

CLASSIFIERS_PYTHON_IMPLEMENTATIONS_TUPLES = tuple((_,) for _ in CLASSIFIERS_PYTHON_IMPLEMENTATIONS)

CLASSIFIERS_VARIOUS_COMBINATIONS = tuple(itertools.combinations(
    CLASSIFIERS_VARIOUS, len(CLASSIFIERS_VARIOUS) - 1)) + (CLASSIFIERS_VARIOUS,)

ALL_CLASSIFIERS_VARIANTS = [
    licenses + versions + implementations + various
    for licenses in CLASSIFIERS_LICENSES_TUPLES
    for versions in CLASSIFIERS_PYTHON_VERSIONS_COMBINATIONS
    for implementations in CLASSIFIERS_PYTHON_IMPLEMENTATIONS_TUPLES
    for various in CLASSIFIERS_VARIOUS_COMBINATIONS]

LINK_EXAMPLES = [
    (None, 'setup.py', True), ('this file', 'setup.py', True), (None, 'test/test_setup.py', True),
    (None, 'test/test_setup.py#L98', True), ('line 5 of this file', 'setup.py#L5', True),
    (None, 'http://site.com', False), (None, '../something/else', False), (None, 'no.thing', False),
    (None, '/my/abs/path', False), ('test dir', 'test', True)]


class UnitTests(unittest.TestCase):
    """Test basic functionalities of the setup boilerplate."""

    def test_find_version(self):
        result = boilerplates.setup.find_version('.', 'setup')
        self.assertIsInstance(result, str)

    def test_find_packages(self):
        results = boilerplates.setup.find_packages()
        self.assertIsInstance(results, list)
        for result in results:
            self.assertIsInstance(result, str)

    def test_requirements(self):
        results = boilerplates.setup.parse_requirements()
        self.assertIsInstance(results, list)
        self.assertTrue(all(isinstance(result, str) for result in results), msg=results)

    def test_requirements_nested(self):
        results = boilerplates.setup.parse_requirements('requirements_test.txt')
        self.assertIsInstance(results, list)
        self.assertTrue(all(isinstance(result, str) for result in results), msg=results)

    def test_requirements_empty(self):
        with tempfile.NamedTemporaryFile('w', delete=False) as reqs_file:
            pass
        results = boilerplates.setup.parse_requirements(reqs_file.name)
        self.assertIsInstance(results, list)
        self.assertEqual(len(results), 0)
        os.remove(reqs_file.name)

    def test_requirements_comments(self):
        reqs = ['# comment', 'numpy', '', '# another comment', 'scipy', '', '# one more comment']
        with tempfile.NamedTemporaryFile('w', delete=False) as reqs_file:
            for req in reqs:
                print(req, file=reqs_file)
        results = boilerplates.setup.parse_requirements(reqs_file.name)
        self.assertIsInstance(results, list)
        self.assertGreater(len(results), 0)
        self.assertLess(len(results), len(reqs))
        os.remove(reqs_file.name)

    def test_python_versions(self):
        for variant in ALL_CLASSIFIERS_VARIANTS:
            with self.subTest(variant=variant):
                result = boilerplates.setup.find_required_python_version(variant)
                if result is not None:
                    self.assertIsInstance(result, str)

    def test_python_versions_combined(self):
        classifiers = [
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.5']
        req = boilerplates.setup.find_required_python_version(classifiers)
        self.assertEqual(req, '>=3.5')

    def test_python_versions_reversed(self):
        classifiers = [
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6']
        req = boilerplates.setup.find_required_python_version(classifiers)
        self.assertEqual(req, '>=3.4')
        req = boilerplates.setup.find_required_python_version(reversed(classifiers))
        self.assertEqual(req, '>=3.4')

    def test_python_versions_none(self):
        result = boilerplates.setup.find_required_python_version([])
        self.assertIsNone(result)

    def test_python_versions_many_only(self):
        classifiers = [
            'Programming Language :: Python :: 2 :: Only',
            'Programming Language :: Python :: 3 :: Only']
        with self.assertRaises(ValueError):
            boilerplates.setup.find_required_python_version(classifiers)

    def test_python_versions_conflict(self):
        classifier_variants = [
            ['Programming Language :: Python :: 2.7',
             'Programming Language :: Python :: 3 :: Only'],
            ['Programming Language :: Python :: 2 :: Only',
             'Programming Language :: Python :: 3.0']]
        for classifiers in classifier_variants:
            with self.assertRaises(ValueError):
                boilerplates.setup.find_required_python_version(classifiers)


class PackageTests(unittest.TestCase):
    """Test methods of Package class."""

    def test_try_fields(self):
        class Package(boilerplates.setup.Package):  # pylint: disable = too-few-public-methods
            name = 'package name'
            description = 'package description'

        self.assertEqual(Package.try_fields('name', 'description'), 'package name')
        self.assertEqual(Package.try_fields('bad_field', 'description'), 'package description')
        with self.assertRaises(AttributeError):
            self.assertIsNone(Package.try_fields())
        with self.assertRaises(AttributeError):
            Package.try_fields('bad_field', 'another_bad_field')

    def test_parse_readme(self):
        class Package(boilerplates.setup.Package):  # pylint: disable = too-few-public-methods
            name = 'package name'
            description = 'package description'
            version = '1.2.3.4'
            url = 'https://github.com/example'

        with tempfile.NamedTemporaryFile('w', suffix='.md', delete=False) as temp_file:
            temp_file.write('test test test')
        result, content_type = Package.parse_readme(temp_file.name)
        os.remove(temp_file.name)
        self.assertIsInstance(result, str)
        self.assertIsInstance(content_type, str)

        prefix = 'https://github.com/example/blob/v1.2.3.4/'
        for name, link, done in LINK_EXAMPLES:
            name = '' if name is None else name + ' '
            text = f'Please see `{name}<{link}>`_ for details.'
            with tempfile.TemporaryDirectory() as temp_folder:
                with tempfile.NamedTemporaryFile(
                        'w', dir=temp_folder, suffix='.rst', delete=False) as temp_file:
                    temp_file.write(text)
                pathlib.Path(temp_folder, 'setup.py').touch()
                pathlib.Path(temp_folder, 'test').mkdir()
                pathlib.Path(temp_folder, 'test', 'test_setup.py').touch()
                result, content_type = Package.parse_readme(temp_file.name)
                os.remove(temp_file.name)
            self.assertIsInstance(result, str)
            self.assertIsInstance(content_type, str)
            if not done:
                self.assertEqual(result, text)
                continue
            if name == '':
                name = link + ' '
            self.assertIn(f'`{name}<{prefix}{link}>`_', result)

    def test_prepare(self):
        version_ = '1.2.3.4.5.6.7'
        long_description_ = 'long package description'

        class Package(boilerplates.setup.Package):
            # pylint: disable = too-few-public-methods, missing-docstring
            name = 'package name'
            version = version_
            description = 'package description'
            long_description = long_description_
            packages = []
            install_requires = []
            python_requires = ''

        self.assertEqual(Package.version, version_)
        self.assertEqual(Package.long_description, long_description_)
        Package.prepare()
        self.assertEqual(Package.version, version_)
        self.assertEqual(Package.long_description, long_description_)

        del Package.long_description
        del Package.packages
        del Package.install_requires
        del Package.python_requires
        Package.prepare()

        del Package.version
        with self.assertRaises(FileNotFoundError):
            Package.prepare()

    def test_prepare_license_files(self):
        # pylint: disable = protected-access

        class Package(boilerplates.setup.Package):
            name = 'package name'
            version = '0.1.0'

        self.assertIsNone(Package._existing_license_file_patterns)
        Package.prepare()
        self.assertIsNotNone(Package._existing_license_file_patterns)
        assert Package._existing_license_file_patterns is not None
        self.assertGreater(len(Package._existing_license_file_patterns), 0)
